/* $Id: common-init.cpp 411 2010-05-06 19:29:33Z oh2gve $
 * 
 * Copyright 2007, 2008, 2009, 2010 Tapio Aaltonen
 * 
 *
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/


// Build config.
#ifdef HAVE_CONFIG_H
#	ifdef __WXMSW__
#		include "../windows/config.h"
#	else
#		include <config.h>
#	endif
#endif

// Own stuff.
#ifdef __WXMSW__
#	include "../windows/main.h"
#else
#	include "main.h"
#endif
#include "kisstnc.h"
#include "uplink.h"
#ifndef __WXMSW__
#	include "ax25port.h"
#endif

// Std stuff.
#include <string>
using std::string;
#include <fstream>
using std::ofstream;
#include <ios>
using std::ios;
#include <sstream>
using std::ostringstream;
#ifndef __WXMSW__
#	include <csignal>
#endif

// Wx stuff.
#include <wx/log.h>


// LOG HANDLING

bool AprsApp::openLog(bool const& initialOpening)
{  
	// Open log file.
	logFile_ = new ofstream(configuration_.logcfg.mainFileName.ToAscii(), ios::app);

	// Check that it opened. Fail if not.
	if ( !logFile_ || !(*logFile_) )
	{  
		wxLogError(wxT("Error: failed to open log file ") + configuration_.logcfg.mainFileName + wxT("."));
		delete logFile_;
		logFile_ = NULL;
		return false;
	}

	// Construct a logging target ant set a nice timestamp format.
	wxLog* wxLogObjectPtr = new wxLogStream(logFile_);
	if ( !wxLogObjectPtr )
	{
		wxLogError(wxT("Error: failed to initialize log file handling."));
		delete logFile_;
		logFile_ = NULL;
		return false;
	}
	wxLogObjectPtr->SetTimestamp(wxT("%Y-%m-%d %H:%M:%S"));
	
	// Set our log as default logging target. This returns pointer to previous
	// target, which must be deleted.
	delete wxLog::SetActiveTarget(wxLogObjectPtr);
	wxLogMessage(wxT("Log file opened."));

#ifndef __WXMSW__
	// If this is first time the log is opened, close stdout and stderr.
	if ( initialOpening )
	{
		if ( dup2(nullFd_, 0) < 0 || dup2(nullFd_, 1) < 0 || dup2(nullFd_, 2) < 0 )
		{  
			// This has never happened.
			wxLogError(wxT("Error: Closing stdio failed."));
			return false;
		}
	}
#endif
	
	return true;
}


void AprsApp::closeLog()
{
	if ( logFile_ != NULL )
	{
		// Remove log file from log target and close it.
		wxLogMessage(wxT("Closing log file.\n"));
		delete wxLog::SetActiveTarget(NULL);   
		delete logFile_;
		logFile_ = NULL;
	}
}


// CONNECTION HANDLING


bool AprsApp::openConnections()
{
	// Construct our login line.
	string greet = "user ";
	greet += configuration_.callsign.ToAscii();
	greet += " pass ";
	greet += aprsPass(configuration_.callsign);
	greet += " vers ";
	greet += PACKAGE;
	greet += " ";
	greet += PACKAGE_VERSION;
	if ( !configuration_.uplink.filter.IsEmpty() )
	{
		greet += " filter ";
		greet += configuration_.uplink.filter.ToAscii();
	}
	greet += "\r\n";

	// First open the server connection.
	wxLogMessage(wxT("Connecting uplink."));

	// Create the thread object.
	uplink_ = new Uplink(configuration_.uplink, greet);
	if ( uplink_->Create() != wxTHREAD_NO_ERROR )
	{  
		delete uplink_;
		uplink_ = 0;
		wxLogError(wxT("Error: cannot create thread for uplink."));
		return false;
	}
	
	// Start the thread.
	uplink_->Run();
	// Wait for it to start up.
	while ( !uplink_->isInitialized() )
	{
		Yield();
	}
	
	// Check that it managed to open it's connection.
	if ( !uplink_->isConnected() )
	{
		wxLogError(wxT("Error: uplink could not be opened."));
		
		// Stop the connection thread, and wait for it to die.
		uplink_->Delete();
		
		delete uplink_;
		uplink_ = NULL;
		return false;
	}
	
	// Uplink successfully opened.
	wxLogMessage(wxT("Uplink connected."));
	

	// Open other connections.

	// Temporary pointer for the connection to be opened.
	Connection* tmpConnPtr = NULL;

	// Loop through all given connection configurations.
	for ( unsigned int i = 0; i < configuration_.connections.size(); i++ )
	{  
		// Tell what we're going to do.
		wxLogMessage(wxString::Format(wxT("Opening connection %d."), i+1));
		
		// Find out what type of connection we are going to open.
		switch ( configuration_.connections.at(i).type )
		{
			
		 case Cfg::AX25PORT:
			{
#ifdef HAVE_LIBAX25
				tmpConnPtr = new Ax25Iface(configuration_.connections.at(i).device, configuration_.connections.at(i).state, configuration_.callsign);
#else
				wxLogError(wxT("Error: AX.25 support is not enabled in this build."));
				return false;
#endif
				break;       
			}
		 case Cfg::SERIALKISSTNC:
			{
				tmpConnPtr = new KissTnc(configuration_.connections.at(i).device, configuration_.connections.at(i).state, configuration_.connections.at(i).portcfg, configuration_.lockdir);
				break;
			}
		 default:
			{
				wxLogMessage(wxString::Format(wxT("Internal error: unknown connection type for connection %d."), i+1));
				return false;
				break;
			}
		}
		
		// Check that port thread got created.
		if ( tmpConnPtr->Create() != wxTHREAD_NO_ERROR )
		{  
			delete tmpConnPtr;
			wxLogError(wxString::Format(wxT("Error: can't create thread for connection %d"), i+1));
			return false;
		}
		
		// Start the thread.
		tmpConnPtr->Run();
		// Wait for it to start up.
		while ( !tmpConnPtr->isInitialized() )
		{
			Yield();
		}
		
		// If we're interested, check wether thread managed to open connection.
		if ( configuration_.connections.at(i).dieOnError && !tmpConnPtr->isConnected() )
		{
			wxLogError(wxString::Format(wxT("Error: connection %d couldn't be opened"), i+1));
		
			// Stop the the connection thread.
			tmpConnPtr->Delete();
			delete tmpConnPtr;
			return false;
		}
		
		// Save the port.
		connections_.push_back(tmpConnPtr);

		// Tell that we were succesfull.
		wxLogMessage(wxString::Format(wxT("Connection %d opened."), i+1));
	}
	
	return true;
}


void AprsApp::closeConnections()
{
	bool didSumthin = !connections_.empty();
	
	// Close connections in descending order.
	for ( int i = connections_.size()-1; i >= 0; --i )
	{
		// Close the connection and wait for it to go.
		connections_.at(i)->Delete();
		delete connections_.at(i);
	}
	
	if ( didSumthin )
	{
		wxLogMessage(wxT("Connections closed."));
	}
	
	// Close uplink.
	if ( uplink_ != 0 )
	{
		uplink_->Delete();
		delete uplink_;
		wxLogMessage(wxT("Uplink closed."));
	}
}


// OTHERS


string AprsApp::aprsPass(wxString callsign)
{
	// Remove possible SSID
	if (callsign.Find('-') != -1)
	{
		callsign = callsign.Mid(0, callsign.Find('-'));
	}
	
	// Convert call to upper case
	callsign = callsign.MakeUpper();
	int result = 0x73e2;
	
	// Go through the whole string and do some XORing.
	for (unsigned int i = 0; i < callsign.Length(); i += 2)
	{
		result ^= (callsign.at(i) << 8);
		if ( i + 1 < callsign.Length() )
		{
			result ^= callsign.at(i + 1);
		}
	}
	
	// Convert result from int to string
	ostringstream output;
	output << (result & 0x7fff);
	
	return output.str();
}
