/* $Id: packetlog.cpp 413 2010-05-09 07:37:22Z oh2gve $
 * 
 * Copyright 2007, 2008, 2009, 2010 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// own stuff.
#include "packetlog.h"
#include "cfg.h"
using Cfg::LogCfg;
#include "logserver.h"
#include "packet.h"
#include "aprspacket.h"

// std stuff.
#include <string>
using std::string;
#include <queue>
using std::queue;
#include <sstream>
using std::ostringstream;
#include <iomanip>

// wx stuff.
#include <wx/string.h>
#include <wx/log.h>
#include <wx/utils.h>
#include <wx/app.h>


Packetlog::Packetlog(LogCfg const& cfg, wxString const& ownCall, AprsPacket::Location const& ownLoc):
initialized_(true), cfg_(cfg), fileNeeded_(!cfg.packetFileName.empty()), file_(NULL), server_(NULL), buffer_()
{
	// Print log message if we're going to start something.
	if ( cfg_.serverNeeded || fileNeeded_ )
	{
		wxLogMessage(wxT("Initializing packet logging."));
	}
	
	// Start the server if needed.
	if ( cfg_.serverNeeded )
	{
		// Create the server thread.
		server_ = new LogServer(cfg.packetSrv, ownCall, ownLoc);
		if ( server_->Create() != wxTHREAD_NO_ERROR )
		{
			wxLogMessage(wxT(" packetlog: Failed to create thread for packet log server."));
			initialized_ = false;
		}
		else
		{
			// Start the server thread.
			server_->Run();
		}
	
		// Make sure server thread has started up.
		if ( initialized_ )
		{
			while ( !server_->isInitialized() )
			{
#ifdef __WXMSW__
				if ( wxTheApp->Pending() )
				{
					wxTheApp->Dispatch();
				}
#endif
				wxTheApp->Yield();
			}
		}
		
		// Check for fatal errors.
		if ( initialized_ && !(*server_) )
		{
			wxLogMessage(wxT(" packetlog: Log server could not be started."));
			initialized_ = false;
		}
	
		// Put note to general log.
		if ( initialized_ )
		{
			wxLogMessage(wxT(" packetlog: Log server started."));
		}
	}
	
	// Open file if needed.
	if ( fileNeeded_ )
	{
		file_ = new Logfile(cfg_.packetFileName.ToAscii());
	
		// Check for fatal errors in file.
		if ( !(*file_) )
		{
			wxLogMessage(wxT(" packetlog: File ") + cfg_.packetFileName + wxT(" could not be opened."));
			initialized_ = false;
		}
	
		// Put note to general log if file was opened.
		if ( initialized_ )
		{
			wxLogMessage(wxT(" packetlog: Log file opened."));
		}
	}
	
	// Tell that we're go.
	if ( initialized_ && (cfg_.serverNeeded || fileNeeded_) )
	{
		wxLogMessage(wxT("Packet logging started."));
	}
}

Packetlog::~Packetlog()
{
	// Close server and file, if they were open.
	if ( server_ != NULL )
	{
		server_->Delete();
		delete server_;
		wxLogMessage(wxT(" packetlog: Log server stopped."));
	}
	if ( file_ != NULL )
	{
		delete file_;
		wxLogMessage(wxT(" packetlog: Log file closed."));
	}
	if ( server_ != NULL || file_ != NULL )
	{
		wxLogMessage(wxT("Packet logging stopped."));
	}
}

bool Packetlog::operator!() const
{
	return !initialized_;
}

bool Packetlog::reopen()
{
	if ( fileNeeded_ )
	{
		delete file_;
		file_ = new Logfile(cfg_.packetFileName.ToAscii());
		if ( !(*file_) )
		{
			wxLogMessage(wxT(" packetlog: Failed to reopen log file."));
			return false;
		}
		else
		{
			wxLogMessage(wxT(" packetlog: Log file reopened."));
		}
	}
	
	return true;
}

void Packetlog::write(Packet* packet, bool wait)
{
	// Check if we're supposed to write later.
	if ( wait )
	{
		buffer_.push(packet);
		return;
	}
	
	// Write this one.
	writeOne(packet);
	
	// Write buffered packets.
	while ( !buffer_.empty() )
	{
		writeOne(buffer_.front());
		buffer_.pop();
	}
}



void Packetlog::writeOne(Packet* packet)
{
	// Here will the log line be created.
	ostringstream commonLine;
	
	// Construct the line.
	commonLine << " i" << packet->input << " o" << packet->output << " " << packet->flag << " " << std::setw(3)
	  << packet->aprsPacket->getTnc2Packet().size() << "B ";
	
	// Write the line.
	if ( file_ != NULL )
	{
		ostringstream line;
		line << packet->timestamp.Format(wxT("%Y-%m-%d %H:%M:%S")).ToAscii();
		line << commonLine.str();
		line << packet->aprsPacket->getLogLine();
		file_->write( line.str() + "\n" );
	}
	if ( server_ != NULL )
	{
		ostringstream line;
		line << packet->timestamp.Format(wxT("%Y-%m-%d %H:%M:%S"), wxDateTime::UTC).ToAscii();
		line << commonLine.str();
		line << packet->aprsPacket->getTnc2Packet();
		server_->write( line.str() + "\r\n" );
	}
	
	// Packet has been logged, so let's erase it.
	delete packet->aprsPacket;
	delete packet->thirdParty;
	delete packet;
}
